package scanner;

import static util.Characters.BACKSLASH;
import static util.Characters.DOUBLE_QUOTE;
import static util.Characters.EOF;
import static util.Characters.HASH;
import static util.Characters.LEFT_PARENTHESIS;
import static util.Characters.LEFT_SQUARE_BRACKET;
import static util.Characters.NEWLINE;
import static util.Characters.PERIOD;
import static util.Characters.RIGHT_PARENTHESIS;
import static util.Characters.RIGHT_SQUARE_BRACKET;
import static util.Characters.SEMICOLON;
import static util.Characters.SINGLE_QUOTE;
import static util.Characters.TICK_MARK;

import java.io.BufferedInputStream;
import java.io.IOException;
import java.io.InputStream;
import java.util.HashMap;
import java.util.Map;

/**
 * Converts a stream of bytes into a stream of Lisp tokens.
 */
public class LispScanner {

    private static Map<Character, Boolean> illegalIdentifierCharacters = new HashMap<>();

    static {
        illegalIdentifierCharacters.put(DOUBLE_QUOTE, true);
        illegalIdentifierCharacters.put(SINGLE_QUOTE, true);
        illegalIdentifierCharacters.put(BACKSLASH, true);
        illegalIdentifierCharacters.put(TICK_MARK, true);
        illegalIdentifierCharacters.put(LEFT_PARENTHESIS, true);
        illegalIdentifierCharacters.put(RIGHT_PARENTHESIS, true);
        illegalIdentifierCharacters.put(LEFT_SQUARE_BRACKET, true);
        illegalIdentifierCharacters.put(RIGHT_SQUARE_BRACKET, true);
        illegalIdentifierCharacters.put(HASH, true);
        illegalIdentifierCharacters.put(PERIOD, true);
        illegalIdentifierCharacters.put(SEMICOLON, true);
    }

    private InputStream inputStream;
    private String inputStreamName;
    private int lineNumber;
    private int columnNumber;

    public LispScanner(InputStream in, String fileName) {
        this.inputStream = new LispFilterInputStream(new BufferedInputStream(in));
        this.inputStreamName = fileName;
        this.lineNumber = 1;
        this.columnNumber = 0;
    }

    public Token nextToken() throws IOException {
        int c;

        while ((c = inputStream.read()) != EOF) {
            char nextChar = (char) c;

            ++columnNumber;

            switch (nextChar) {
            case NEWLINE:
                moveToNewLine();
                break;
            case LEFT_PARENTHESIS:
                return new Token(Token.Type.LEFT_PAREN, "(", inputStreamName, lineNumber, columnNumber);
            case RIGHT_PARENTHESIS:
                return new Token(Token.Type.RIGHT_PAREN, ")", inputStreamName, lineNumber, columnNumber);
            case SINGLE_QUOTE:
                return new Token(Token.Type.QUOTE_MARK, "\'", inputStreamName, lineNumber, columnNumber);
            case DOUBLE_QUOTE:
                return retrieveString(nextChar);
            default:
                if (Character.isWhitespace(nextChar)) {
                    continue;
                } else if (Character.isDigit(nextChar)) {
                    return retrieveNumber(nextChar);
                } else if (isLegalIdentifierCharacter(nextChar)) {
                    return retrieveIdentifier(nextChar);
                } else {
                    throw new RuntimeException("illegal character " + "\'" + nextChar + "\'" + " - line " + lineNumber
                            + " column " + columnNumber);
                }
            }
        }

        return new Token(Token.Type.EOF, "EOF", inputStreamName, lineNumber, columnNumber);
    }

    private Token retrieveString(char firstDoubleQuote) throws IOException {
        StringBuffer text = new StringBuffer();
        int startLine = lineNumber;
        int startColumn = columnNumber;
        char prevChar = firstDoubleQuote;

        text.append(firstDoubleQuote);

        int c;

        while ((c = inputStream.read()) != EOF) {
            char nextChar = (char) c;

            ++columnNumber;
            text.append(nextChar);

            switch (nextChar) {
            case NEWLINE:
                moveToNewLine();
                break;
            case DOUBLE_QUOTE:
                if (prevChar != BACKSLASH) {
                    // we have found the terminating double quote

                    return new Token(Token.Type.STRING, text.toString(), inputStreamName, startLine, startColumn);
                }

                // this is an escaped double quote
            }

            prevChar = nextChar;
        }

        // the end of 'inStream' was reached before the terminating double
        // quote

        throw new RuntimeException("unterminated quoted string" + " - line " + startLine + " column " + startColumn);
    }

    private Token retrieveNumber(char firstDigit) throws IOException {
        StringBuffer text = new StringBuffer();
        int startColumn = columnNumber;

        text.append(firstDigit);
        inputStream.mark(1);

        int c;

        while ((c = inputStream.read()) != EOF) {
            char nextChar = (char) c;

            if (Character.isDigit(nextChar)) {
                // 'nextChar' is a digit in this number

                text.append(nextChar);
                ++columnNumber;
            } else {
                // we have reached the end of the number

                inputStream.reset(); // unread the last character

                return new Token(Token.Type.NUMBER, text.toString(), inputStreamName, lineNumber, startColumn);
            }

            inputStream.mark(1);
        }

        // there are no more bytes to be read from 'inStream' after this number
        // token

        return new Token(Token.Type.NUMBER, text.toString(), inputStreamName, lineNumber, startColumn);
    }

    private Token retrieveIdentifier(char firstChar) throws IOException {
        StringBuffer text = new StringBuffer();
        int startColumn = columnNumber;

        text.append(firstChar);
        inputStream.mark(1);

        int c;

        while ((c = inputStream.read()) != EOF) {
            char nextChar = (char) c;

            if (isLegalIdentifierCharacter(nextChar)) {
                // 'nextChar' is part of the identifier

                text.append(nextChar);
                ++columnNumber;
            } else {
                // we have reached the end of this identifier

                inputStream.reset(); // unread the last character

                return new Token(Token.Type.IDENTIFIER, text.toString(), inputStreamName, lineNumber, startColumn);
            }

            inputStream.mark(1);
        }

        // there are no more bytes to be read from 'inStream' after this
        // identifier token

        return new Token(Token.Type.IDENTIFIER, text.toString(), inputStreamName, lineNumber, startColumn);
    }

    private void moveToNewLine() {
        lineNumber++;
        columnNumber = 0;
    }

    private boolean isLegalIdentifierCharacter(char c) {
        return (!Character.isWhitespace(c)) && (illegalIdentifierCharacters.get(c) == null);
    }

}
